/**~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Universidad de los Andes (Bogot - Colombia)
 * Departamento de Ingeniera de Sistemas y Computacin 
 * Licenciado bajo el esquema Academic Free License version 2.1 
 *
 * Proyecto Cupi2 (http://cupi2.uniandes.edu.co)
 * Ejercicio: cupiTaxonomia
 * Autor: Equipo Cupi2 2015
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 */

package uniandes.cupi2.cupiTaxonomia.mundo;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;

import estructuras.arboles.ArbolBO;
import estructuras.arboles.ArbolN;
import estructuras.arboles.NodoAN;

/**
 * Clase que representa el rbol taxonmico. <br>
 * <b> inv: </b> <br>
 * taxonRaiz != null.
 */
public class ArbolTaxonomico implements Serializable
{
    // -------------------------------------------------------------
    // Atributos
    // -------------------------------------------------------------

    /**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	/**
     * Taxn raz del rbol taxonmico.
     */
    
    private ArbolBO<String, SerVivo> seres;
    
    private ArbolN<String, Taxon> taxonomia;



    // -------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------

    /**
     * Constructor del rbol taxonmico.<br>
     * <b>post: </b> El atributo archivoTaxonomia se inicializa con la ruta de archivo dada por parmetro. <br>
     * Si existe el archivo en la ruta dada por parmetro, se crea la clase a partir la informacin en ese archivo. <br>
     * Si el archivo no existe, se inicializa el taxn raz. <br>
     * @param pRutaArchivo La ruta del archivo del cual se carga la informacin del rbol taxonmico. pRutaArchivo != null && pRutaArchivo != "".
     * @throws Exception Si no se puede abrir o leer el archivo con el estado del rbol taxonmico.
     */
    public ArbolTaxonomico(  ) 
    {
    	Taxon raiz = new Taxon();
    	taxonomia = new ArbolN<String, Taxon>(new NodoAN<String, Taxon>(raiz.darNombre(), raiz));
    	seres = new ArbolBO<String, SerVivo>();

        
    }

    // -------------------------------------------------------------
    // Mtodos
    // -------------------------------------------------------------

    /**
     * Retorna el nmero de taxones del rbol taxonmico. <br>
     * @return Nmero de taxones del rbol taxonmico.
     */
    public int darNumTaxones( )
    {
    	return taxonomia.size();
    }

    /**
     * Retorna el nmero de seres vivos del rbol taxonmico. <br>
     * @return Nmero de seres vivos del rbol taxonmico.
     */
    public int darNumSeresVivos( )
    {
    	return seres.size();
    }

    /**
     * Retorna la lista de seres vivos del taxn con el nombre y tipo dado. <br>
     * <b> pre: </b> El taxn con el nombre y el tipo dado existen.
     * @param pTipoTaxon Tipo del taxn. pTipoTaxon pertenece a {LUCA, DOMINIO, REINO, FILO, CLASE, ORDEN, FAMILIA, GENERO, ESPECIE}.
     * @param pNombreTaxon Nombre del taxn. pNombreTaxon != null y pNombreTaxon != "".
     * @return Lista de seres vivos que pertenecen a ese taxn.
     */
    public Collection<SerVivo> darSeresVivos( String pNombreTaxon )
    {
    	Collection<SerVivo> pListaSeres = new ArrayList<SerVivo>();
    	NodoAN<String, Taxon> n = taxonomia.getNodo(pNombreTaxon);
    	Collection<Taxon> taxonesH = new ArrayList<Taxon>();
    	n.darHojas(taxonesH);
    	
    	for (Taxon taxon : taxonesH)
    	{
    		if(taxon instanceof Especie )
    		{
    			Especie esp = (Especie)taxon;
    			if(esp.darSerVivo() != null)
    			{    			
    				pListaSeres.add(esp.darSerVivo());
    			}
    		}
		}
    	return pListaSeres;
    }

    /**
     * Retorna el taxn raz del rbol taxonmico.
     * @return Taxn raz del rbol taxonmico.
     */
    public NodoAN<String, Taxon> darTaxonRaiz( )
    {
        return taxonomia.darRaiz();
    }


    /**
     * Retorna la lista con los nombres de los taxones del tipo dado.
     * @param pTipoTaxon Tipo de los taxones. pTipoTaxon pertenece a {LUCA, DOMINIO, REINO, FILO, CLASE, ORDEN, FAMILIA, GENERO, ESPECIE}.
     * @return Lista con los nombres de los taxones del tipo dado.
     */
    public Collection<Taxon> darTaxonesTipo( int pTipoTaxon )
    {
        return taxonomia.valuesLevel(pTipoTaxon);
    }

    /**
     * Retorna la lista de taxones del rbol taxonmico. Los taxones de agregan en preorden. <br>
     * @return Lista de taxones del rbol taxonmico en preorden.
     */
    public Collection<Taxon> darTaxonesPreorden( )
    {
    	return taxonomia.values();
    }

    /**
     * Retorna la lista de taxones del rbol taxonmico Los taxones de agregan en postorden. <br>
     * @return Lista de taxones del rbol taxonmico en postorden.
     */
    public Collection<Taxon> darTaxonesPostorden( )
    {
        return  taxonomia.valuesPos();
    }

    /**
     * Agrega un ser vivo a un taxn del rbol taxonmico. <br>
     * <b> pre: </b> El taxn existe y es de tipo ESPECIE. <br>
     * <b> post: </b> Se ha agregado un nuevo ser vivo a un taxn del rbol taxonmico.
     * @param pNombreTaxonPadre Nombre del taxn padre al que se va a agregar el ser vivo. pNombreTaxon != null && pNombreTaxon != "".
     * @param pNombreComun Nombre comn del ser vivo a agregar. pNombreComun != null && pNombreComun != "".
     * @param pNombreCientifico Nombre cientfico del ser vivo a agregar. pNombreCientifico != null && pNombreCientifico != "".
     * @param pCaracteristicas Caractersticas del ser vivo a agregar. pCaracteristicas != null && pCaracteristicas != "".
     * @param pImagen Ruta con la imagen del ser vivo a agregar. pImagen != null y pImagen != "".
     * @throws Exception Si ya existe un ser vivo con el nombre cientfico dado.
     * @throws Exception Si el taxn ya tiene un ser vivo.
     */
    public void agregarSerVivo( String pNombreTaxonPadre, String pNombreComun, String pNombreCientifico, String pCaracteristicas, String pImagen ) throws Exception
    {
      if(seres.containsKey(pNombreCientifico))
      {
    	  throw new Exception("Ya existe el animal");
      }
      SerVivo s = new SerVivo(pNombreComun, pNombreCientifico, pCaracteristicas, pImagen);
      seres.put(pNombreCientifico, s);
      taxonomia.put(pNombreTaxonPadre, pNombreCientifico, new Especie(pNombreCientifico, s));
    }

    /**
     * Agrega un nuevo sub-taxn al taxn con el nombre dado. <br>
     * <p> pre: </b> El tipo del taxn hijo es mayor al tipo del taxn padre. <br>
     * <b> post: </b> Se agreg un nuevo taxn al rbol taxonmico.
     * @param pNombreTaxonPadre Nombre del taxn padre del taxn a agregar. pNombreTaxonPadre != null && pNombreTaxonPadre != "".
     * @param pTipoTaxonHijo Tipo del taxn a agregar. pTipoTaxonHijo pertenece a {LUCA, DOMINIO, REINO, FILO, CLASE, ORDEN, FAMILIA, GENERO, ESPECIE}.
     * @param pNombreTaxonHijo Nombre del taxn a agregar. pNombreTaxonHijo != null && pNombreTaxonHijo != "".
     * @throws Exception Si ya existe un taxn con el nombre dado.
     */
    public void agregarTaxon( String pNombreTaxonPadre, TipoTaxon pTipoTaxonHijo, String pNombreTaxonHijo ) throws Exception
    {
        taxonomia.put(pNombreTaxonPadre, pNombreTaxonHijo, new Taxon(pTipoTaxonHijo, pNombreTaxonHijo));

    }
    
    public Collection<SerVivo> darSeresVivosNombreMenor(String nombre)
    {
    	return seres.headMap(nombre).values();
    }

    /**
     * Elimina el taxn con el nombre y el tipo dado del rbol taxonmico. <br>
     * <b> pre: </b> El taxn a eliminar existe. <br>
     * <b> post: </b> Se elimin el taxn del rbol taxonmico.
     * @param pTipoTaxon Tipo del taxn a eliminar. pTipoTaxon pertenece a {LUCA, DOMINIO, REINO, FILO, CLASE, ORDEN, FAMILIA, GENERO, ESPECIE}.
     * @param pNombreTaxon Nombre del taxn a eliminar. pNombreTaxon != null && pNombreTaxon != "".
     */
    public void eliminarTaxon( Taxon taxon )
    {
    	taxonomia.remove(taxon.darNombre());
    	if(taxon instanceof Especie)
    	{
    		SerVivo ser = ((Especie)taxon).darSerVivo();
    		if(ser != null)
    		{
    			seres.remove(ser.darNombreCientifico());
    		}
    	}
    }

    /**
     * 
     * @param nombre
     * @return
     */
    public Collection<SerVivo> darMenores(String nombre)
    {
    	return seres.headMap(nombre).values();
    }
    
    /**
     * 
     * @param nombre
     * @return
     */
    public Collection<SerVivo> darMayoresIguales(String nombre)
    {
    	return seres.tailMap(nombre).values();
    }
    
    
    public Collection<SerVivo> darEntre(String nombreInf, String nombreSup)
    {
    	return seres.subMap(nombreInf, nombreSup).values();
    }

    // -----------------------------------------------------------------
    // Puntos de Extensin
    // -----------------------------------------------------------------

    /**
     * Mtodo para la extensin 1.
     * @return respuesta1.
     */
    public String metodo1( )
    {
        return "Respuesta 1.";
    }

    /**
     * Mtodo para la extensin2.
     * @return respuesta2.
     */
    public String metodo2( )
    {
        return "Respuesta 2.";
    }
}